//Class Genetic Pathway
import java.io.*;
import java.util.*;
import java.lang.*;

class GenPathway
{	
	//for original pathway
	Vector vctGeneSet;	//the gene set vector
	Vector vctRelationship;	//the relationship vector
	Vector vctImpact;	//positive--"+" or negative--"-" impact to the vertual sink node
	String key;	//used to prevent duplicated genetic pathways
	double score;	//genetic pathway score (1 - p)

	//for computable pathway
	Vector vctCompRelationship;	//vector for computable relationships
	
	public GenPathway (String strRegulator, String strRegulatee, String strType) throws Exception
	{
		//initilization
		vctGeneSet = new Vector();
		vctRelationship = new Vector();
		key = "";
		
		//assign a new relationship
		Relationship rltTemp = new Relationship();
		rltTemp.strFirstGene = strRegulator;
		rltTemp.strSecondGene = strRegulatee;
		rltTemp.strType = strType;
		
		//add to vector
		vctGeneSet.add(strRegulator);
		vctGeneSet.add(strRegulatee);
		vctRelationship.add(rltTemp);

		//set the key to delete the duplicated pathways
		for (int i =0; i<vctGeneSet.size() ; i++)
		{
			key += (String)vctGeneSet.get(i);
		}
	}

	public GenPathway (GenPathway firstMember, GenPathway secondMember) throws Exception
	{
		//initilization
		vctGeneSet = new Vector();
		vctRelationship = new Vector();
		key = "";

		//merge the two genetic pathways
		for (int i = 0;i<firstMember.vctGeneSet.size() ;i++ )
		{
			vctGeneSet.add((String)firstMember.vctGeneSet.get(i));
		}
		for (int i = 1;i<secondMember.vctGeneSet.size() ;i++ )
		{
			vctGeneSet.add((String)secondMember.vctGeneSet.get(i));
		}
		for (int i = 0;i<firstMember.vctRelationship.size() ;i++ )
		{
			vctRelationship.add((Relationship)firstMember.vctRelationship.get(i));
		}
		for (int i = 0;i<secondMember.vctRelationship.size() ;i++ )
		{
			vctRelationship.add((Relationship)secondMember.vctRelationship.get(i));
		}

		//set the key to delete the duplicated pathways
		for (int i =0; i<vctGeneSet.size() ; i++)
		{
			key += (String)vctGeneSet.get(i);
		}
	}

	public boolean getComputablePathway(GeneExpression gxpExpression) throws Exception
	{
		//initilization
		vctCompRelationship = new Vector();
		vctImpact = new Vector();

		//assign impact for each gene
		String strCurrentImpact = "+";
		for (int i = vctRelationship.size()-1;i>=0 ;i--)
		{
			vctImpact.insertElementAt(strCurrentImpact,0);	
			if (((Relationship)vctRelationship.get(i)).strType.equals("$negative"))
			{
				if (strCurrentImpact.equals("+"))
				{
					strCurrentImpact = "-";
				}
				else
				{
					strCurrentImpact = "+";
				}
			}
		}
		vctImpact.insertElementAt(strCurrentImpact,0);

		//find the first observable gene
		int intFirstPosition = 0;
		for (int i = 0;i<vctGeneSet.size() ;i++ )
		{
			intFirstPosition = i;
			if (gxpExpression.hshGeneNameToRef.get((String)vctGeneSet.get(i))!=null)
			{	
				break;
			}
		}
		
		//find the last observable gene
		int intLastPosition = 0;
		for (int i = vctGeneSet.size()-1; i>=0;i-- )
		{
			intLastPosition = i;
			if (gxpExpression.hshGeneNameToRef.get((String)vctGeneSet.get(i))!=null)
			{
				break;
			}
		}

		if (intFirstPosition>=intLastPosition)
		{
			vctCompRelationship.removeAllElements();
			return false;
		}

		//from the first observable gene to the last observable gene
		int intStart = intFirstPosition;
		int intEnd = intFirstPosition+1;
		String strCurrentType = "$positive";
		for (;intEnd<=intLastPosition ;intEnd++)
		{
			//merge relationships with unobservable gene
			if (((Relationship)vctRelationship.get(intEnd-1)).strType.equals("$negative"))
			{
				if (strCurrentType.equals("$positive"))
				{
					strCurrentType = "$negative";
				}
				else
				{
					strCurrentType = "$positive";
				}
			}

			if (gxpExpression.hshGeneNameToRef.get((String)vctGeneSet.get(intEnd))!=null)
			{
				//add to computable relationship set
				Relationship rltTemp = new Relationship();
				rltTemp.strFirstGene = (String)vctGeneSet.get(intStart);
				rltTemp.strSecondGene = (String)vctGeneSet.get(intEnd);
				rltTemp.strType = strCurrentType;
				vctCompRelationship.add(rltTemp); 
				intStart = intEnd;
				strCurrentType = "$positive";
			}
		}
		return true;
	}

	public double getPathwayZScore(GeneExpression gxpExpression) throws Exception
	{
		//initialization
		double dblPathwayZScore = 0;
		
		//if the pathway is not computable
		if (vctCompRelationship.size()==0)
		{
			return -1000;	//return a significant negative value
		}

		//for each computable relationship
		for (int i = 0;i<vctCompRelationship.size() ;i++ )
		{
			//compute its correlation
			double dblCorrelation = gxpExpression.getCorrelation(((Relationship)vctCompRelationship.get(i)).strFirstGene, ((Relationship)vctCompRelationship.get(i)).strSecondGene);

			//get the z-score of the correlation
			double dblZScore = gxpExpression.getZScore(dblCorrelation);
			//add up to the pathway correlation score w.r.t. the regulation type
			if (((Relationship)vctCompRelationship.get(i)).strType.equals("$positive"))
			{
				//if a positive correlation, then add z-score of the relationship
				dblPathwayZScore += dblZScore;
			}
			else
			{
				//if a negative correlation, then add minus of z-score of the relationship
				dblPathwayZScore -= dblZScore;
			}
		}
		dblPathwayZScore *= 1/(Math.sqrt(vctCompRelationship.size()));
		return dblPathwayZScore;
	}

	public boolean getScore(double dblPathwayZScore, double[] dblZScoreDist) throws Exception
	{
		//if the pathway is not computable
		if (dblPathwayZScore < -500)
		{
			score = 0.0001;
			return false;
		}
		
		//if the pathway z-score is large than the first value
		if (dblPathwayZScore > dblZScoreDist[9999])
		{
			score = 0.9999;
			return true;
		}
		//if the pathway z-score falls in the distribution
		for (int i = 9999;i>0 ;i-- )
		{
			if (dblPathwayZScore <= dblZScoreDist[i] && dblPathwayZScore > dblZScoreDist[i-1])
			{
				score = ((double)i)/10000.0;
				return true;
			}
		}
		//else
		score = 0.0001;
		return false;
	}
};

class Relationship
{
	String strFirstGene;
	String strSecondGene;
	String strType;
};