//Class Gene Expression
import java.io.*;
import java.util.*;
import java.lang.*;

class GeneExpression
{
	String[] strGeneName;
	HashMap hshGeneNameToRef;
	double[][] dblValue;
	double[][] dblReValue;

	double dblArrayCorrMean;
	double dblArrayCorrDev;
	
	public GeneExpression (String strFileName) throws Exception
	{
		int intTemporal = 0;
		int intGeneCount = 0;
		Vector vctReadInBuffer = new Vector();

		//read in gene expression files and save them in buffer
		BufferedReader brInFile = new BufferedReader(new FileReader("expression\\"+strFileName));
		String strDataEntry = brInFile.readLine();
		while (strDataEntry!=null)
		{
			vctReadInBuffer.add(strDataEntry);
			strDataEntry = brInFile.readLine();
		}
		StringTokenizer tknToken = new StringTokenizer((String)vctReadInBuffer.get(0),",");
		tknToken.nextToken();
		while (tknToken.hasMoreTokens())
		{
			intTemporal++;
			tknToken.nextToken();
		}
		
		//initialize expression and RE datasets
		strGeneName = new String[vctReadInBuffer.size()];
		hshGeneNameToRef = new HashMap();
		dblValue = new double[vctReadInBuffer.size()][intTemporal];
		dblReValue = new double[vctReadInBuffer.size()][intTemporal];
		
		//parse expression and get RE values
		for (int i = 0;i<vctReadInBuffer.size() ;i++ )
		{
			tknToken = new StringTokenizer((String)vctReadInBuffer.get(i),",");
			strGeneName[i] = tknToken.nextToken();
			hshGeneNameToRef.put((String)strGeneName[i], i);
			for (int j = 0;j<intTemporal ;j++ )
			{
				dblValue[i][j] = Double.parseDouble(tknToken.nextToken());
				if (dblValue[i][j] >= dblValue[i][0])
				{
					dblReValue[i][j] = dblValue[i][j]/dblValue[i][0]-1;
				}
				else
				{
					dblReValue[i][j] = 1 - dblValue[i][0]/dblValue[i][j];
				}
			}
		}

		//estimate the mean and devidation of correlation
		getBackground();
	}

	public void getBackground() throws Exception
	{
		//initialization
		dblArrayCorrMean = 0;
		dblArrayCorrDev = 0;
		double[] dblCorrArray = new double[strGeneName.length*strGeneName.length];
		
		//for each gene pair compute their correlation and save in an array
		int intArrayCursor = 0;
		for (int i = 0;i<strGeneName.length ;i++ )
		{
			for (int j = 0;j<strGeneName.length ;j++ )
			{
				dblCorrArray[intArrayCursor] = getCorrelation(strGeneName[i], strGeneName[j]);
				intArrayCursor++;
			}
		}

		//compute the mean and devidation respectively
		for (int i = 0;i<intArrayCursor ;i++ )
		{
			dblArrayCorrMean += dblCorrArray[i];
		}
		dblArrayCorrMean /= intArrayCursor;
		for (int i = 0;i<intArrayCursor ;i++ )
		{
			dblArrayCorrDev += Math.pow(dblCorrArray[i]-dblArrayCorrMean,2);
		}
		dblArrayCorrDev = Math.sqrt(dblArrayCorrDev/intArrayCursor);
	}

	public double getCorrelation(String strFirstGene, String strSecondGene) throws Exception
	{
		//initialization
		double dblCorrelation = 0;
		int intTemporal = 0;
		int intFirstGeneRef = (Integer)hshGeneNameToRef.get(strFirstGene);
		int intSecondGeneRef = (Integer)hshGeneNameToRef.get(strSecondGene);

		//get temporal count
		intTemporal = dblReValue[0].length;

		//if temporal count is bigger than 2
		if (intTemporal>2)
		{
			dblCorrelation = getMultiCorr(dblReValue[intFirstGeneRef], dblReValue[intSecondGeneRef], intTemporal);
		}
		//if temporal count equals to 2
		else
		{
			dblCorrelation = getBinCorr(dblReValue[intFirstGeneRef], dblReValue[intSecondGeneRef]);
		}
		
		return dblCorrelation;
	}

	public double getZScore(double dblCorrelation) throws Exception
	{
		double dblZScore = (dblCorrelation-dblArrayCorrMean)/dblArrayCorrDev;
		return dblZScore;
	}

	public double getMultiCorr(double[] a, double[] b, int intTemporal)
	{
		double mean1 = 0, mean2 = 0, dev1 = 0, dev2 = 0, correlation = 0;
		
		for (int i = 0;i<intTemporal ;i++ )
		{
			mean1 += a[i];
			mean2 += b[i];
		}
		mean1 /= intTemporal;
		mean2 /= intTemporal;

		for (int i = 0;i<intTemporal ;i++ )
		{
			dev1 += Math.pow(a[i]-mean1,2);
			dev2 += Math.pow(b[i]-mean2,2);
		}
		dev1 = Math.sqrt(dev1/intTemporal);
		dev2 = Math.sqrt(dev2/intTemporal);
		if (dev1==0)
		{
			dev1 = 0.01;
		}
		if (dev2==0)
		{
			dev2 = 0.01;
		}

		for (int i = 0;i<intTemporal;i++ )
		{
			correlation += (a[i]-mean1)*(b[i]-mean2);
		}
		correlation /= intTemporal*dev1*dev2;

		return correlation;
	}

	public double getBinCorr(double[] a, double[] b)
	{
		double correlation = 0;
		if (a[1] == 0)
		{
			a[1] = 0.01;
		}
		if (b[1] == 0)
		{
			b[1] = 0.01;
		}

		if (Math.abs(a[1])>Math.abs(b[1]))
		{
			correlation = b[1]/a[1];
		}
		else
		{
			correlation = a[1]/b[1];
		}

		return correlation;
	}
	
	public void printRE() throws Exception
	{
		for (int i =0;i<dblReValue.length ;i++ )
		{
			for (int j = 0;j<dblReValue[1].length ;j++ )
			{
				System.out.print(dblReValue[i][j]+" ");
			}
			System.out.println(" ");
		}
	}
};