//the main function of pathway identification system
import java.io.*;
import java.util.*;
import java.lang.*;

class Main
{
	public static void main (String[] args) throws Exception
	{
		boolean blnTestMode = false;

		//parse parameters
		blnTestMode = false;
		double dblPThreshold = 0;
		double dblFdrThreshold = 0;
		if (args.length == 0)
		{
			System.out.println("Threshold for statistical significance checking is required!");	
			System.exit(1);
		}
		for (int i=0;i<args.length ;i++ )
		{
			StringTokenizer tknParaFactor = new StringTokenizer (args[i], " =,");
			String strParaEntry = tknParaFactor.nextToken(); 
			if (strParaEntry.equals("p")||strParaEntry.equals("P"))
			{
				dblPThreshold = Double.parseDouble(tknParaFactor.nextToken());				
			}
			else
				dblFdrThreshold = Double.parseDouble(tknParaFactor.nextToken());				
		}
		if (blnTestMode)
		{
			System.out.println("P = " + dblPThreshold);
			System.out.println("FDR = " + dblFdrThreshold);
		}
		blnTestMode = false;

		//get gene expression files and save in an array
		blnTestMode = false;
		File flExpressionFolder = new File("expression");
		File[] flExpression = flExpressionFolder.listFiles();
		if (blnTestMode)
		{
			System.out.println(flExpression.length);
		}
		blnTestMode = false;

		//get signaling pathway files and save in an array
		blnTestMode = false;
		File flPathwayFolder = new File("pathway");
		File[] flPathway = flPathwayFolder.listFiles();
		if (blnTestMode)
		{
			System.out.println(flPathway.length);
		}
		blnTestMode = false;
		
		//for each gene expression file 
		for (int i = 0;i<flExpression.length ;i++ )
		{
			//preprocessing gene expression data
			blnTestMode = false;
			GeneExpression gxpExpression = new GeneExpression(flExpression[i].getName());
			if (blnTestMode)
			{
				gxpExpression.printRE();
			}
			blnTestMode = false;
			
			//for each signaling pathway
			for (int j = 0;j<flPathway.length ;j++ )
			{
				//extracting genetic relationships
				blnTestMode = false;
				SigPathway spthPathway = new SigPathway(flPathway[j].getName());
				if (blnTestMode)
				{
					spthPathway.printRegulation();
				}
				blnTestMode = false;

				//searching and scoring computable genetic pathways
				blnTestMode = false;
				spthPathway.geneticPathwayAnalysis(gxpExpression);
				if (blnTestMode)
				{
					spthPathway.printScore();
				}
				blnTestMode = false;
				//generating hypotheses 
				spthPathway.hypothesesGeneration(gxpExpression, flPathway[j].getName().substring(0,flPathway[j].getName().length()-4), flExpression[i].getName().substring(0,flExpression[i].getName().length()-4), dblPThreshold, dblFdrThreshold);
			}
			System.out.println(flExpression[i].getName().substring(0,flExpression[i].getName().length()-4)+" finished");
		}
	}
};