import java.io.*;
import java.util.*;
import java.lang.*;

class SigPathway
{	
	String[] strProteinName;	//the array used to keep protein name
	HashMap hshProteinNameToRef;	//given a protein name, return its array reference
	Protein[] prtProtein;	//the array used to keep protein, the order should be identical with the order of protein name
	Interaction[] itrInteraction; //the array used to keep the interaction
	Vector vctGenPathway;	//each signaling pathway is associated with a set of genetic pathways
	double[][] dblZScoreDist = new double[100][10000];	//background distribution of genetic pathway z-score with observalbe genes from 2 to 99
	
	public SigPathway (String strFileName) throws Exception
	{
		Vector vctRegulationBuffer = new Vector();
		Vector vctAssociationBuffer = new Vector();
		
		//read in pathway files and save them in buffer
		BufferedReader brInFile = new BufferedReader(new FileReader("pathway\\"+strFileName));
		String strDataEntry = brInFile.readLine();
		strDataEntry = brInFile.readLine();
		while (strDataEntry!=null&&(!strDataEntry.equals("Association:")))
		{
			vctRegulationBuffer.add(strDataEntry);
			strDataEntry = brInFile.readLine();
		}
		strDataEntry = brInFile.readLine();
		while (strDataEntry!=null)
		{
			vctAssociationBuffer.add(strDataEntry);
			strDataEntry = brInFile.readLine();
		}

		//initialization
		strProteinName = new String[vctAssociationBuffer.size()];
		hshProteinNameToRef = new HashMap();
		prtProtein = new Protein[vctAssociationBuffer.size()];
		itrInteraction = new Interaction[vctRegulationBuffer.size()];
		vctGenPathway = new Vector();

		//Processing association
		for (int i = 0;i<prtProtein.length ;i++ )
		{
			StringTokenizer tknToken = new StringTokenizer((String)vctAssociationBuffer.get(i));
			Protein prtTemp = new Protein();
			prtTemp.strName = tknToken.nextToken();
			strProteinName[i] = prtTemp.strName;
			hshProteinNameToRef.put((String)strProteinName[i],i);
			String strTemp = tknToken.nextToken();
			while (tknToken.hasMoreTokens())
			{
				strTemp += ","+tknToken.nextToken();
			}
			prtTemp.strAssoGene = strTemp;
			prtProtein[i] = prtTemp;
		}

		//Processing regulation
		for (int i = 0;i<itrInteraction.length ;i++ )
		{
			StringTokenizer tknToken = new StringTokenizer((String)vctRegulationBuffer.get(i));
			Interaction itrTemp = new Interaction();
			itrTemp.intFirstRef = (Integer)hshProteinNameToRef.get(tknToken.nextToken());
			itrTemp.intSecondRef = (Integer)hshProteinNameToRef.get(tknToken.nextToken());
			itrTemp.strType = tknToken.nextToken();
			itrInteraction[i] = itrTemp;
		}
	}

	public void geneticPathwayAnalysis(GeneExpression gxpExpression) throws Exception
	{
		//genetic relationship extraction
		Vector vctBinRelationship = new Vector();	//a vector used to hold binary relationships
		for (int i =0 ;i<itrInteraction.length ;i++ )
		{
			StringTokenizer tknFirstToken = new StringTokenizer (prtProtein[itrInteraction[i].intFirstRef].strAssoGene, ",");
			while (tknFirstToken.hasMoreTokens())
			{
				String strFirstToken = tknFirstToken.nextToken();
				StringTokenizer tknSecondToken = new StringTokenizer (prtProtein[itrInteraction[i].intSecondRef].strAssoGene, ",");
				while (tknSecondToken.hasMoreTokens())
				{
					String strSecondToken = tknSecondToken.nextToken();
					GenPathway gpthPathway;
					if (itrInteraction[i].strType.equals("$stimulation"))
					{
						gpthPathway = new GenPathway (strFirstToken, strSecondToken, "$positive");
					}
					else
					{
						gpthPathway = new GenPathway (strFirstToken, strSecondToken, "$negative");
					}
					vctBinRelationship.add(gpthPathway);
				}
			}
		}

		//connect genetic relationship to get genetic pathways
		//initialize a work set for genetic pathways
		Vector vctWorkSet = new Vector();
		for (int i =0;i<vctBinRelationship.size() ;i++ )
		{
			vctWorkSet.add((GenPathway)vctBinRelationship.get(i));
		}
		boolean blnIsChange;
		do
		{
			blnIsChange = false;

			//remove all elements in genetic pathway set
			vctGenPathway = new Vector();	

			//copy from workset back to genetic pathway set
			for (int i = 0;i<vctWorkSet.size() ;i++ )
			{
				vctGenPathway.add((GenPathway)vctWorkSet.get(i));
			}

			//initilize a new workset
			vctWorkSet = new Vector();

			//for each genetic pathway in the pathway set
			for (int i =0 ;i<vctGenPathway.size() ;i++ )
			{
				boolean blnIsExtendable = false;
				//for each pathway from binary relationship vector
				for (int j = 0;j<vctBinRelationship.size() ;j++ )
				{
					if (((String)((GenPathway)vctGenPathway.get(i)).vctGeneSet.firstElement()).equals(((String)((GenPathway)vctBinRelationship.get(j)).vctGeneSet.lastElement())))
					{
						GenPathway gpthPathway = new GenPathway((GenPathway)vctBinRelationship.get(j),(GenPathway)vctGenPathway.get(i));
						vctWorkSet.add(gpthPathway);
						blnIsExtendable = true;
						blnIsChange = true;
					}
					if (((String)((GenPathway)vctGenPathway.get(i)).vctGeneSet.lastElement()).equals(((String)((GenPathway)vctBinRelationship.get(j)).vctGeneSet.firstElement())))
					{
						GenPathway gpthPathway = new GenPathway((GenPathway)vctGenPathway.get(i),(GenPathway)vctBinRelationship.get(j));
						vctWorkSet.add(gpthPathway);						
						blnIsExtendable = true;
						blnIsChange = true;
					}
				}
				if (!blnIsExtendable)
				{
					vctWorkSet.add((GenPathway)vctGenPathway.get(i));
				}
			}

			//delete duplicates in work set
			HashMap hshTemp = new HashMap();
			for (int i =0;i<vctWorkSet.size() ;i++ )
			{
				if (hshTemp.get(((GenPathway)vctWorkSet.get(i)).key)==null)
				{
					hshTemp.put(((GenPathway)vctWorkSet.get(i)).key, 1);
				}
				else
				{
					vctWorkSet.remove(i);
					i--;
				}
			}
		}
		//while the pathway number in the original set equals to that in the working set
		//while (vctGenPathway.size()!=vctWorkSet.size());
		while (blnIsChange);

		//for each genetic pathway
		for (int i = 0;i<vctGenPathway.size() ;i++ )
		{
			//get its computable genetic pathway
			((GenPathway)vctGenPathway.get(i)).getComputablePathway(gxpExpression);

			//generate background z-score distribution for the computable genetic pathway
			if (((GenPathway)vctGenPathway.get(i)).vctCompRelationship.size()>0)
			{
				getZScoreBackground(((GenPathway)vctGenPathway.get(i)).vctCompRelationship.size()+1, gxpExpression);
			}
			
			//compute genetic pathway score
			//get the z-score of a genetic pathway
			double dblPathwayZScore = ((GenPathway)vctGenPathway.get(i)).getPathwayZScore(gxpExpression);

			//estimate the significance of that pathway
			((GenPathway)vctGenPathway.get(i)).getScore(dblPathwayZScore, dblZScoreDist[((GenPathway)vctGenPathway.get(i)).vctCompRelationship.size()+1]);
		}
	}

	public void hypothesesGeneration(GeneExpression gxpExpression, String strPathwayName, String strSampleName, double dblPThreshold, double dblFdrThreshold) throws Exception
	{	
		//perform p-value cutOff
		if (dblPThreshold>0)
		{
			double dblCutOff = 1-dblPThreshold;
			for (int i =0;i<vctGenPathway.size() ;i++ )
			{
				//only reserve pathways with score exceeds the threshold
				if (((GenPathway)vctGenPathway.get(i)).score<dblCutOff)
				{
					vctGenPathway.remove(i);
					i--;
				}
			}
		}
		
		//perform fdr cutOff
		if (dblFdrThreshold>0)
		{
			//given the q-value, to generate the cutOff for FDR
			//extract and rank the score
			double[] dblScoreArray = new double[vctGenPathway.size()];
			for (int i =0;i<vctGenPathway.size() ;i++ )
			{
				dblScoreArray[i] = 1-((GenPathway)vctGenPathway.get(i)).score;
			}
			Arrays.sort(dblScoreArray);
			//compute the cn w.r.t dependent test
			/*double cn = 0;
			for (int i =1;i<=vctGenPathway.size() ;i++ )
			{
				cn += 1.0/i;
			}*/
			double cn = 1;
			double dblCutOff = 0;
			for (int i=vctGenPathway.size()-1;i>=0;i-- )
			{
				if (dblScoreArray[i]<(i*dblFdrThreshold)/(cn*vctGenPathway.size()))
				{
					dblCutOff = dblScoreArray[i]; 
				}
			}
			//conserve at least one result
			if (dblCutOff==0&&vctGenPathway.size()>0)
			{
				dblCutOff = dblScoreArray[0];
			}
			dblCutOff = 1-dblCutOff;
			for (int i =0;i<vctGenPathway.size() ;i++ )
			{
				//only reserve pathways with score exceeds the threshold
				if (((GenPathway)vctGenPathway.get(i)).score<dblCutOff)
				{
					vctGenPathway.remove(i);
					i--;
				}
			}
		}

		//output genetic pathway hypotheses
		File folder = new File("result\\"+strSampleName);
		folder.mkdir();
		String strGenFileName = strPathwayName+"_genetic.txt";
		FileWriter fstream = new FileWriter("result\\"+strSampleName+"\\"+strGenFileName);
		BufferedWriter out = new BufferedWriter(fstream);
		out.write(strPathwayName+", genetic hypotheses:\n");
		for (int i = 0;i<vctGenPathway.size() ;i++ )
		{
			out.write((String)((GenPathway)vctGenPathway.get(i)).vctGeneSet.get(0));
			for (int j = 0;j<((GenPathway)vctGenPathway.get(i)).vctRelationship.size() ;j++ )
			{
				if (((Relationship)(((GenPathway)vctGenPathway.get(i)).vctRelationship.get(j))).strType.equals("$positive"))
				{
					out.write("-->");
				}
				else
					out.write("--|");
				out.write(((Relationship)(((GenPathway)vctGenPathway.get(i)).vctRelationship.get(j))).strSecondGene);
			}
			out.write("\t"+ ((GenPathway)vctGenPathway.get(i)).score +"\n");
		}
		out.close();

		//output signaling pathway hypotheses
		String strSigFileName = strPathwayName+"_signaling.txt";
		fstream = new FileWriter("result\\"+strSampleName+"\\"+strSigFileName);
		out = new BufferedWriter(fstream);
		double[] dblStatus = new double[gxpExpression.dblValue[0].length];//aggregated RE values of genes on the pathway
		double dblConf = 0;//confidence score for a signaling pathway
		double dblTemp = 0;
		for (int i = 0;i<vctGenPathway.size() ;i++ )
		{
			dblTemp += ((GenPathway)vctGenPathway.get(i)).score;
		}
		for (int i = 0;i<vctGenPathway.size() ;i++ )
		{
			dblConf += ((GenPathway)vctGenPathway.get(i)).score*((GenPathway)vctGenPathway.get(i)).score/dblTemp;
		}
		for (int i = 0;i<gxpExpression.dblValue[0].length ;i++ )
		{
			dblStatus[i] = 0;
			for (int j = 0;j<vctGenPathway.size() ;j++ )
			{
				for (int k = 0;k<((GenPathway)vctGenPathway.get(j)).vctGeneSet.size() ;k++ )
				{
					String strGeneName = (String)(((GenPathway)vctGenPathway.get(j)).vctGeneSet.get(k));
					if (gxpExpression.hshGeneNameToRef.get(strGeneName)!=null)
					{
						double dblGeneCount = ((double)(((GenPathway)vctGenPathway.get(j)).vctCompRelationship.size()+1));
						double dblReValue = gxpExpression.dblReValue[(Integer)(gxpExpression.hshGeneNameToRef.get(strGeneName))][i];
						double dblImpact;
						if (((String)(((GenPathway)vctGenPathway.get(j)).vctImpact.get(k))).equals("+"))
						{
							dblImpact = 1.0;
						}
						else
						{
							dblImpact = -1.0;
						}
						dblStatus[i] += dblImpact/dblGeneCount * dblReValue *((GenPathway)vctGenPathway.get(j)).score/dblTemp;
					}			
				}
			}
		}

		out.write(strPathwayName+", status:\n");
		for (int i = 0;i<gxpExpression.dblValue[0].length ;i++ )
		{
			out.write(dblStatus[i]+"\t");
		}
		out.write("\nConfidence:"+dblConf);
		out.close();
	}

	public boolean getZScoreBackground(int intSize, GeneExpression gxpExpression) throws Exception
	{
			
		//if distribution already exists
		if (dblZScoreDist[intSize][9999]!=0)
		{
			return false;
		}
		
		//initialization
		Random generator = new Random();
		double[] dblCurrentDist = new double[10000];

		//for 10000 times
		for (int i = 0;i<10000 ;i++ )
		{
			//randomly generate a genetic pathway with intSize genes
			GenPathway gpthPathway = null;
			
			//randomly generate an active gene
			String strActiveGene = gxpExpression.strGeneName[generator.nextInt(gxpExpression.strGeneName.length)];

			//generate the next intSize-1 genes
			for (int j = 0;j<intSize-1 ;j++ )
			{
				//randomly generate a gene
				String strCurrentGene = gxpExpression.strGeneName[generator.nextInt(gxpExpression.strGeneName.length)];

				//randomly generate a relationship type
				String strType;
				int intType = generator.nextInt(2);
				if (intType==0)
				{
					strType = "$positive";
				}
				else
				{
					strType = "$negative";
				}

				//generate a genetic pathway with the active gene, current generated gene and relationship type
				GenPathway gpthCurrentPathway = new GenPathway(strActiveGene, strCurrentGene, strType);

				//connect the previous gpthPathway with the current generated pathway
				if (gpthPathway == null)
				{
					gpthPathway = gpthCurrentPathway;
				}
				else
				{
					GenPathway gpthNewPathway = new GenPathway(gpthPathway, gpthCurrentPathway);
					gpthPathway = gpthNewPathway;
				}

				//set current gene to be the active gene
				strActiveGene = strCurrentGene;
			}

			//compute the pathway z-score for the generated genetic pathway
			gpthPathway.getComputablePathway(gxpExpression);
			dblCurrentDist[i] = gpthPathway.getPathwayZScore(gxpExpression);
		}

		//rank the array
		Arrays.sort(dblCurrentDist);
		dblZScoreDist[intSize] = dblCurrentDist;
		return true;
	}
	
	public void printAssociation()
	{
		for (int i = 0;i<prtProtein.length ;i++ )
		{
			System.out.println(prtProtein[i].strName+" "+prtProtein[i].strAssoGene);
		}	
	}

	public void printRegulation()
	{
		for (int i = 0;i<itrInteraction.length ;i++ )
		{
			System.out.println(prtProtein[itrInteraction[i].intFirstRef].strName+" "+prtProtein[itrInteraction[i].intSecondRef].strName+" "+itrInteraction[i].strType);
		}
	}

	public void printScore()
	{
		for (int i = 0;i<vctGenPathway.size() ;i++ )
		{
			System.out.println(((GenPathway)vctGenPathway.get(i)).vctCompRelationship.size()+" "+((GenPathway)vctGenPathway.get(i)).key+" "+((GenPathway)vctGenPathway.get(i)).score);
		}
	}
};

class Protein
{
	String strName;
	String strAssoGene;
};

class Interaction
{
	int intFirstRef;
	int intSecondRef;
	String strType;
};